# logs repo/commit info

puts "ENV['GITHUB_WORKFLOW_REF'] #{ENV['GITHUB_WORKFLOW_REF']}\n" \
     "ENV['GITHUB_WORKFLOW_SHA'] #{ENV['GITHUB_WORKFLOW_SHA']}\n" \
     "ENV['GITHUB_REPOSITORY']   #{ENV['GITHUB_REPOSITORY']}\n"   \
     "ENV['GITHUB_REF_TYPE']     #{ENV['GITHUB_REF_TYPE']}\n"     \
     "ENV['GITHUB_REF']          #{ENV['GITHUB_REF']}\n"          \
     "ENV['GITHUB_REF_NAME']     #{ENV['GITHUB_REF_NAME']}"
